-- Version 1.1
-- 6 March 2023

TAG_SHOE =		"trolleyShoe"
TAG_NODE =		"trolleyNode"
TAG_GROUP =		"trolleyGroup"
TAG_SWITCH =	"trolleySwitch"
TAG_IGNORE =	"trolleyIgnore"
TAG_CONNECT =	"trolleyConnect"

POLE_MAX = 		0.9
ANGLE_MIN =		2.5

PARENT_NODE = 	"trolleybus."

DEFAULT_KEY_SWITCH =	"backspace"
DEFAULT_KEY_CONNECT =	"t"
DEFAULT_KEY_CONFIRM =	"return"

function init()
	DEBUG = GetBoolParam("DEBUG", false)

	switchToggleKey =	GetString(PARENT_NODE.."key.switch")
	connectToggleKey =	GetString(PARENT_NODE.."key.connect")
	confirmKey =		GetString(PARENT_NODE.."key.confirm")
	switchToggleKey =	switchToggleKey ~= ""	and switchToggleKey or	DEFAULT_KEY_SWITCH
	connectToggleKey =	connectToggleKey ~= ""	and connectToggleKey or	DEFAULT_KEY_CONNECT
	confirmKey =		confirmKey ~= ""		and confirmKey or		DEFAULT_KEY_CONFIRM

	keySet = ""
	preConnect = false
	preCount = 0
	preReady = false
	preDone = false
	prePage = 1
	preSelect = 0
	preFlip = false
	connect = false
	doneCount = 0
	initDone = false
	switchCount = 0
	switchShow = false
	switchLock = false
	switchCheck = false
	switchNearby = false
	switchCtrl = false
	angleCount = 0
	timer = 0
	broken = false

	vehiSelf = FindVehicle()
	vehiShoes = FindShapes(TAG_SHOE)
	for i, shape in ipairs(vehiShoes) do
		SetShapeCollisionFilter(shape, 1, 255-1)
		local joints = GetShapeJoints(shape)
		local shapeT = GetShapeWorldTransform(shape)
		local bodyPos = GetBodyTransform(GetShapeBody(shape)).pos
		local poleL = VecLength(VecSub(GetShapeCenter(shape), bodyPos))
		local tempJoint, conNode
		vehiShoes[i] = {shoe = shape, length = poleL, node = 0, neighbour = {}, joints = {}}
		for j, joint in ipairs(joints) do
			if HasTag(joint, TAG_CONNECT) and tempJoint ~= 0 then
				tempJoint = joint
				conNode = GetJointOtherShape(joint, shape)
				vehiShoes[i].node = conNode
				connect = true
			end
			Delete(joint)
		end
	end
end

function tick()
	if broken then return end
	local newKeySet = GetString(PARENT_NODE.."initDone")
	if newKeySet ~= "" and keySet ~= newKeySet then
		NEWswitchToggleKey =	GetString(PARENT_NODE.."key.switch")
		NEWconnectToggleKey =	GetString(PARENT_NODE.."key.connect")
		NEWconfirmKey =		GetString(PARENT_NODE.."key.confirm")
		switchToggleKey =	NEWswitchToggleKey ~= ""	and NEWswitchToggleKey or	DEFAULT_KEY_SWITCH
		connectToggleKey =	NEWconnectToggleKey ~= ""	and NEWconnectToggleKey or	DEFAULT_KEY_CONNECT
		confirmKey =		NEWconfirmKey ~= ""			and NEWconfirmKey or		DEFAULT_KEY_CONFIRM
		keySet = newKeySet
	end
	if preConnect and not preReady then
		preCount = 0
		for i, list in ipairs(vehiShoes) do
			vehiShoes[i].node = 0
			vehiShoes[i].neighbour = {}
			vehiShoes[i].joints = {}
			local shape = list.shoe
			local matchList = {}
			if not HasTag(shape, TAG_CONNECT) then SetTag(shape, TAG_CONNECT) end
			local nodeStr = GetTagValue(shape, TAG_CONNECT)
			if nodeStr ~= "" then
				nodeStr:gsub("_(%d+)", function(s) table.insert(matchList, tonumber(s) or 0) end)
				vehiShoes[i].match = matchList
				preCount = preCount+1
			end
		end
		preReady = preCount == #vehiShoes
	elseif preConnect and preReady then
		for i, list in ipairs(vehiShoes) do
			local shape = list.shoe
			local nodeStr = GetTagValue(shape, TAG_CONNECT)
			preReady = nodeStr ~= ""
		end
	end

	if GetPlayerVehicle() ~= vehiSelf then return end
	if InputPressed(connectToggleKey) then
		if connect then
			connect = false
			preConnect = false
			switchLock = false
			switchNearby = false
		else
			connect = preConnect and preDone
			preConnect = not preConnect
			preReady = false
			preDone = false
			prePage = 1
		end
		initDone = false
		doneCount = 0
	end
	if InputPressed(switchToggleKey) and not switchLock then
		if connect then switchCtrl = not switchCtrl else preFlip = not preFlip end
	end
end

function draw(dt)
	if GetBool(PARENT_NODE.."inOption") then return end
	if broken then return end
	if GetPlayerVehicle() ~= vehiSelf then return end
	if preConnect and preReady then
		local selected = false
		local shape = vehiShoes[prePage].shoe
		local node = vehiShoes[prePage].node
		local matchList = vehiShoes[prePage].match

		for p, point in ipairs(matchList) do
			local shapePos = GetShapeCenter(shape)
			local nodePos = GetShapeBottom(point)
			if node == point then
				DrawLine(shapePos, nodePos, 0, 1, 0)
			else
				DrawLine(shapePos, nodePos)
			end
			local midPoint = VecLerp(shapePos, nodePos, 0.5)
			local screenX, screenY, screenDist = UiWorldToPixel(midPoint)
			if screenDist > 0 then
				UiPush()
					UiTranslate(screenX, screenY)
					UiAlign("center middle")
					UiButtonImageBox("ui/common/box-solid-6.png", 5, 5, 0.1, 0.1, 0.1, 0.75)
					UiFont("regular.ttf", 24)
					UiTextButton(node == point and string.upper(confirmKey) or p)
				UiPop()
			end
			if InputPressed(p) and not selected then vehiShoes[prePage].node = point end
		end

		if node ~= 0 and InputPressed(confirmKey) then
			connect = prePage == #vehiShoes
			preDone = connect
			preConnect = not connect
			prePage = prePage+1
		end
	end
	UiPush()
		UiTranslate(UiCenter(), UiHeight()-86)
		UiAlign("center middle")
		UiButtonImageBox("ui/common/box-solid-6.png", 5, 5, 0.1, 0.1, 0.1, 0.75)
		UiFont("regular.ttf", 24)
		if not switchNearby then
			switchLock = false
			if preConnect then
				timer = timer+dt*2.5
				local dotMatch = {".", "..", "..."}
				local dotStr = dotMatch[math.floor(timer)%3+1]
				UiTextButton(preReady and ("Connecting"..dotStr) or "Find a capturer")
			else
				UiTextButton(string.upper(connectToggleKey).." to "..(connect and "disconnect" or "connect (when possible)"))
			end
		else
			if not switchLock then
				UiTextButton(string.upper(switchToggleKey).." to toggle")
			else
				UiTextButton(string.upper(connectToggleKey).." to "..(connect and "disconnect" or "connect (when possible)"))
				UiColor(1, 0.55, 0)
				UiFont("bold.ttf", 24)
			end
			UiTranslate(0, -36)
			UiTextButton(switchCtrl and "Turnout" or "Through")
		end
	UiPop()
end

function update()
	if broken then return end
	if not connect then return end
	for i, list in ipairs(vehiShoes) do
		local shoe = list.shoe
		local poleMax = list.length
		local node = list.node
		local neighbour = list.neighbour
		local checkJoints = list.joints
		local shapePos = GetShapeCenter(shoe)
		local poleBody = GetShapeBody(shoe)

		if IsBodyBroken(poleBody) then broken = true break end
		for j, joint in ipairs(checkJoints) do
			if IsJointBroken(joint) then connect = false break end
		end
		if not connect then break end

		local bodyPos = GetBodyTransform(poleBody).pos
		local nodePos = GetShapeBottom(node)
		local nodeDist = VecLength(VecSub(shapePos, nodePos))
		local centreVec = VecSub(bodyPos, nodePos)
		local centreDist = VecLength(centreVec)
		local newMin, newNode = nodeDist, node
		local shoeNodeVec = VecSub(nodePos, shapePos)
		local changeMin, changeNode = -1, 0

		if DEBUG then
			DrawShapeOutline(node, 1)
			DebugLine(bodyPos, nodePos, 0, 1, 1)
		end

		if HasTag(shoe, TAG_CONNECT) then RemoveTag(shoe, TAG_CONNECT) end

		for n, nNode in ipairs(neighbour) do
			local nNodePos = GetShapeBottom(nNode)
			local neighDist = VecLength(VecSub(shapePos, nNodePos))
			local shoeNeighVec = VecSub(nNodePos, shapePos)
			local neighLen = VecLength(VecSub(VecNormalize(shoeNeighVec), VecNormalize(shoeNodeVec)))

			if DEBUG then DebugLine(bodyPos, nNodePos, 0, 1, 0) end

			if newMin > neighDist then newMin, newNode = neighDist, nNode end
			if neighLen > changeMin then changeMin, changeNode = neighLen, nNode end
		end
		
		if #neighbour == 0 or newNode ~= node then
			switchCheck = true
			vehiShoes[i].node = newNode
			local useSwitch = HasTag(newNode, TAG_SWITCH)
			local switchRaw = GetTagValue(newNode, TAG_SWITCH)
			useSwitch = switchRaw ~= ""
			local switchPatten = "([^%-]+)%-+([^%-]+)"
			local switchGroup = {switchRaw:gsub(switchPatten, "%1"), switchRaw:gsub(switchPatten, "%2")}
			local jointGroup = useSwitch and (switchCtrl and switchGroup[2] or switchGroup[1]) or vehiShoes[i].prevGroup
			local noGroup = 0
			switchLock = useSwitch or switchLock
			SetShapeCollisionFilter(newNode, 1, 255-1)
			vehiShoes[i].neighbour = {}
			vehiShoes[i].joints = {}
			for j, joint in ipairs(GetShapeJoints(newNode)) do
				local haveGroup = HasTag(joint, TAG_GROUP)
				local locGroup = GetTagValue(joint, TAG_GROUP)
				jointGroup = jointGroup or (haveGroup and locGroup) or jointGroup

				if not HasTag(joint, TAG_IGNORE) then
					if haveGroup then
						if string.find(locGroup, jointGroup) then
							local otherShape = GetJointOtherShape(joint, newNode)
							if HasTag(otherShape, TAG_NODE) then table.insert(vehiShoes[i].neighbour, otherShape) end
						end
					else
						local otherShape = GetJointOtherShape(joint, newNode)
						if HasTag(otherShape, TAG_NODE) then
							table.insert(vehiShoes[i].neighbour, otherShape)
							noGroup = noGroup+1
						end
					end
				end
			end
			if noGroup == #vehiShoes[i].neighbour then vehiShoes[i].prevGroup = nil else vehiShoes[i].prevGroup = jointGroup end

			if #vehiShoes[i].neighbour > 2 then
				local checkNeighbour = {}
				for n, nNode in ipairs(vehiShoes[i].neighbour) do
					if #checkNeighbour < 2 then
						useSwitch = useSwitch or HasTag(nNode, TAG_SWITCH)
						table.insert(checkNeighbour, nNode)
					end
				end
				vehiShoes[i].neighbour = checkNeighbour
			else
				for n, nNode in ipairs(vehiShoes[i].neighbour) do
					useSwitch = useSwitch or HasTag(nNode, TAG_SWITCH)
				end
			end

			for j, joint in ipairs(GetShapeJoints(newNode)) do
				local haveGroup = HasTag(joint, TAG_GROUP)
				local locGroup = GetTagValue(joint, TAG_GROUP)
				if not HasTag(joint, TAG_IGNORE) then
					if haveGroup then
						if string.find(locGroup, jointGroup) then
							local otherShape = GetJointOtherShape(joint, newNode)
							if HasTag(otherShape, TAG_NODE) then table.insert(vehiShoes[i].joints, joint) end
						end
					else
						local otherShape = GetJointOtherShape(joint, newNode)
						if HasTag(otherShape, TAG_NODE) then table.insert(vehiShoes[i].joints, joint) end
					end
				end
			end
			switchCount = switchCount+(useSwitch and 1 or 0)
		else
			local cableNodePos = GetShapeBottom(changeNode)
			local cableFunc = {pos = nodePos, dir = VecNormalize(VecSub(cableNodePos, nodePos))}
			local radAngle = VecGetAngleBetween(cableFunc.dir, centreVec, true)
			if radAngle == radAngle then
				local closetDist = math.cos(radAngle)*centreDist
				local perpenDist = math.sin(radAngle)*centreDist
				local closetPoint = VecAdd(cableFunc.pos, VecScale(cableFunc.dir, closetDist))
				local shoeSqrt = math.sqrt(poleMax^2-perpenDist^2)
				if shoeSqrt == shoeSqrt then
					local shoeDist = closetDist > shoeSqrt and closetDist-shoeSqrt or closetDist+shoeSqrt
					local shoePoint = VecAdd(cableFunc.pos, VecScale(cableFunc.dir, shoeDist))

					if DEBUG then
						DrawLine(nodePos, cableNodePos)
						DebugLine(bodyPos, closetPoint)
						DebugLine(bodyPos, shoePoint, 1, 0, 0)
					end

					if not initDone then
						if VecLength(VecSub(shoePoint, shapePos)) < 0.25 then doneCount = doneCount + 1 end
						ConstrainPosition(0, poleBody, shoePoint, shapePos, 5, 100)
					else
						if perpenDist > poleMax*POLE_MAX then connect = false break
						elseif VecLength(VecSub(shoePoint, shapePos)) > 0.5 then connect = false break
						elseif VecLength(VecSub(nodePos, bodyPos)) > poleMax and #neighbour < 2 then connect = false break
						else ConstrainPosition(0, poleBody, shoePoint, shapePos, 50, 500) end
					end
				end
			end
		end
	end
	if switchCheck then switchNearby = switchCount > 0 end
	initDone = doneCount == #vehiShoes
	switchCount = 0
	switchCheck = false
end

function GetShapeCenter(shape)
	local min, max = GetShapeBounds(shape)
	return VecLerp(min, max, 0.5)
end

function GetShapeBottom(shape)
	local x, y = GetShapeSize(shape)
	local shapeT = GetShapeWorldTransform(shape)
	return TransformToParentPoint(shapeT, Vec(x/20, 0, y/20))
end

function VecGetAngleBetween(A, B, useArg)
	local factor1 = VecDot(A, B)
	local factor2 = VecLength(A)*VecLength(B)
	local argAngle = math.acos(factor1/factor2)
	return useArg and argAngle or argAngle*180/math.pi
end

function HandleNaN(value, default)
	return value == value and value or default
end